/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core.nodetype;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.NamespaceRegistry;
import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;
import javax.jcr.ValueFactory;
import javax.jcr.nodetype.NoSuchNodeTypeException;
import javax.jcr.nodetype.NodeType;
import javax.jcr.nodetype.NodeTypeIterator;
import javax.jcr.nodetype.NodeTypeManager;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlPolicy;
import org.exoplatform.services.jcr.config.RepositoryEntry;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeTypeManager;
import org.exoplatform.services.jcr.core.nodetype.NodeTypeValue;
import org.exoplatform.services.jcr.core.nodetype.NodeTypeValuesList;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.impl.core.LocationFactory;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.util.EntityCollection;
import org.exoplatform.services.log.ExoLogger;
import org.jibx.runtime.BindingDirectory;
import org.jibx.runtime.IBindingFactory;
import org.jibx.runtime.IUnmarshallingContext;
import org.jibx.runtime.JiBXException;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: NodeTypeManagerImpl.java 6332 2006-06-20 13:12:00Z peterit $
 */
public class NodeTypeManagerImpl implements ExtendedNodeTypeManager {
  
  protected static Log log = ExoLogger.getLogger("jcr.NodeTypeManagerImpl");

  private static final String NODETYPES_FILE = "nodetypes.xml";

  public static final String NT_ROOT = "/jcr:system/jcr:nodetypes";

  private ValueFactory valueFactory;

  private LocationFactory locationFactory;

  private List<NodeType> nodeTypes;

  private String accessControlPolicy;
  
  //private DataManager dataManager = null;
  
  private NamespaceRegistry namespaceRegistry = null;
  
  private NodeTypeDataPersister persister = null;

  public NodeTypeManagerImpl(RepositoryEntry config, 
      //DataManager dataManager, 
      LocationFactory locationFactory, 
      ValueFactoryImpl valueFactory, 
      NamespaceRegistry namespaceRegistry, 
      NodeTypeDataPersister persister) 
      throws RepositoryException, InstantiationException {
    this(
        //dataManager, 
        locationFactory, 
        valueFactory, 
        namespaceRegistry, 
        config.getAccessControl(), 
        persister, 
        new ArrayList <NodeType>());
    initDefault();
  }
  
  protected NodeTypeManagerImpl(
      //DataManager dataManager,
      LocationFactory locationFactory, 
      ValueFactoryImpl valueFactory,
      NamespaceRegistry namespaceRegistry,
      String accessControlPolicy,
      NodeTypeDataPersister persister,
      List <NodeType>  nodeTypes) {
    this.nodeTypes = nodeTypes; 
    this.valueFactory = valueFactory;
    this.locationFactory = locationFactory;
    this.namespaceRegistry = namespaceRegistry;
    this.accessControlPolicy = accessControlPolicy;
    //this.dataManager = dataManager;
    this.persister = persister;
    
//    this.persister = new NodeTypeDataPersister(dataManager, /*this,*/ valueFactory, namespaceRegistry); 
  }
  
  public WorkspaceNTManagerImpl createWorkspaceNTManager(SessionImpl session) 
    throws RepositoryException {
    WorkspaceNTManagerImpl wntm = new WorkspaceNTManagerImpl(
        //dataManager, 
        namespaceRegistry,
        accessControlPolicy, 
        session,
        nodeTypes);
    return wntm;
  }

  /**
   * 6.10.4 Returns the NodeType specified by nodeTypeName. If no node type by
   * that name is registered, a NoSuchNodeTypeException is thrown.
   */
  public NodeType getNodeType(String nodeTypeName) throws NoSuchNodeTypeException, RepositoryException {
    return getNodeType(locationFactory.parseJCRName(nodeTypeName).getInternalName());
  }
  
  public NodeType getNodeType(InternalQName qName) throws NoSuchNodeTypeException, RepositoryException {
    NodeTypeImpl nt = findNodeType(qName);
    if (nt == null)
      throw new NoSuchNodeTypeException("NodeTypeManager.getNodeType(): NodeType '" 
          + qName.getAsString() + "' not found.");
    else
      return nt;
  }
  
  public NodeTypeImpl findNodeType(InternalQName qName) throws RepositoryException {
    for(int i=0; i<nodeTypes.size(); i++) {
      NodeTypeImpl nt = (NodeTypeImpl)nodeTypes.get(i);
      if(nt.getQName().equals(qName)) {
        return nt;
      }
    }
    return null;
  }
  
  /** 6.10.4 Returns all available node types. */
  public NodeTypeIterator getAllNodeTypes() {

    EntityCollection ec = new EntityCollection();
    ec.addAll(nodeTypes);
    return ec;
  }

  /**
   * Returns an iterator over all available primary node types.
   * 
   * @return An <code>NodeTypeIterator</code>.
   * @throws RepositoryException if an error occurs.
   */
  public NodeTypeIterator getPrimaryNodeTypes() throws RepositoryException {
    EntityCollection ec = new EntityCollection();
    NodeTypeIterator allTypes = getAllNodeTypes();
    while (allTypes.hasNext()) {
      NodeType type = allTypes.nextNodeType();
      if (!type.isMixin())
        ec.add(type);
    }
    return ec;
  }

  /**
   * Returns an iterator over all available mixin node types.
   * 
   * @return An <code>NodeTypeIterator</code>.
   * @throws RepositoryException if an error occurs.
   */
  public NodeTypeIterator getMixinNodeTypes() throws RepositoryException {
    EntityCollection ec = new EntityCollection();
    NodeTypeIterator allTypes = getAllNodeTypes();
    while (allTypes.hasNext()) {
      NodeType type = allTypes.nextNodeType();
      if (type.isMixin())
        ec.add(type);
    }
    return ec;
  }

  /**
   * Registers node type from object
   * 
   * @param nodeType - the node type object
   * @param alreadyExistsBehaviour - if node type with such a name already
   *          exists: IGNORE_IF_EXISTS - does not register new node (default)
   *          FAIL_IF_EXISTS - throws RepositoryException REPLACE_IF_EXISTS -
   *          replaces registerd type with new one
   * @throws RepositoryException
   */
  public void registerNodeType(NodeType nodeType, int alreadyExistsBehaviour)
      throws RepositoryException {

    if (nodeType == null) {
      throw new RepositoryException("NodeType object " + nodeType 
          + " is null");
    }
    
    long start = System.currentTimeMillis();   

    if (accessControlPolicy.equals(AccessControlPolicy.DISABLE)
        && nodeType.getName().equals("exo:accessControllable")) {
      throw new RepositoryException("NodeType exo:accessControllable is DISABLED");
    }
    
    InternalQName qname = (nodeType instanceof NodeTypeImpl ? ((NodeTypeImpl) nodeType).getQName() : 
      (nodeType instanceof WorkspaceNTImpl ? ((WorkspaceNTImpl) nodeType).getQName() : null));
    if (qname == null) {
      throw new RepositoryException("NodeType implementation class " + nodeType.getClass().getName() 
          + " is not supported in this method");
    }
    
    if (findNodeType(qname) != null) {
      if (alreadyExistsBehaviour == FAIL_IF_EXISTS) {
        throw new RepositoryException("NodeType " + nodeType.getName() + " is already registered");
      }
      return;
    }
    
    nodeTypes.add(nodeType);
    
    if (persister.isPersisted()) {
      try {
        if (!persister.hasNodeTypeData(nodeType.getName())) {
          persister.addNodeType(nodeType);
          persister.saveChanges();
        }
      } catch(InvalidItemStateException e) {
        log.warn("Error of storing node type " + nodeType.getName() + ". May be node type already registered .");
      }
      log.info("NodeType " + nodeType.getName() + " initialized. " + (System.currentTimeMillis() - start) + " ms");
    } else {
      log.debug("NodeType " + nodeType.getName() + " registered but not initialized (storage is not initialized). " + (System.currentTimeMillis() - start) + " ms");
      //log.debug("Can't persist node type " + nodeType.getName() + ". Persistent storage is not initialized.");
    }    
  }

  /**
   * Registers node type from class containing the NT definition The class
   * should have constructor with one parameter NodeTypeManager
   * 
   * @param nodeTypeType - Class containing node type definition
   * @param alreadyExistsBehaviour if node type with such a name already exists:
   *          IGNORE_IF_EXISTS - does not register new node (default)
   *          FAIL_IF_EXISTS - throws RepositoryException REPLACE_IF_EXISTS -
   *          replaces registerd type with new one
   * @throws RepositoryException
   */
  public void registerNodeType(Class nodeTypeType, int alreadyExistsBehaviour)
      throws RepositoryException, InstantiationException {

    registerNodeType(makeNtFromClass(nodeTypeType), alreadyExistsBehaviour);

  }

  public void registerNodeType(NodeTypeValue nodeTypeValue, int alreadyExistsBehaviour)
      throws RepositoryException {

    if (accessControlPolicy.equals(AccessControlPolicy.DISABLE)
        && nodeTypeValue.getName().equals("exo:accessControllable")) {
      // skip this node, so it's not necessary at this runtime
      //String msg = "Skiping node '" + nodeTypeValue.getName()
      //    + "' -- it's not necessary at this runtime";
      log.warn("Node type exo:accessControllable is not register due to DISABLE control policy");
      return;
    }

    // We have to validate node value before registering it
    nodeTypeValue.validateNodeType();
    NodeTypeImpl nodeType = new NodeTypeImpl(this, nodeTypeValue);
    registerNodeType(nodeType, alreadyExistsBehaviour);

  }

  /**
   * Registers node type using xml document
   * 
   * @param xml
   * @param alreadyExistsBehaviour
   * @throws RepositoryException
   * @see org.exoplatform.services.jcr.core.nodetype.ExtendedNodeTypeManager#registerNodeType(java.io.InputStream,
   *      int)
   */
  public void registerNodeTypes(InputStream xml, int alreadyExistsBehaviour)
      throws RepositoryException {

    try {
      IBindingFactory factory = BindingDirectory.getFactory(NodeTypeValuesList.class);
      IUnmarshallingContext uctx = factory.createUnmarshallingContext();
      NodeTypeValuesList nodeTypeValuesList = (NodeTypeValuesList) uctx.unmarshalDocument(xml, null);
      ArrayList ntvList = nodeTypeValuesList.getNodeTypeValuesList();

      long start = System.currentTimeMillis();
      for (int i = 0; i < ntvList.size(); i++) {
        if (ntvList.get(i) != null) {
          NodeTypeValue nodeTypeValue = (NodeTypeValue) ntvList.get(i);
          registerNodeType(nodeTypeValue, alreadyExistsBehaviour);
        } else {
          // Hm! Smth is wrong in xml document
          log.error("Empty nodeTypeValue in xml document, index: " + i + ", skiping...");
        }
      }
      if (System.currentTimeMillis() - start > 5)
        log.info("Nodetypes registered from xml definitions (count: " + ntvList.size() 
            + "). " + (System.currentTimeMillis() - start) + " ms.");
    } catch (JiBXException e) {
      throw new RepositoryException("Error in config initialization " + e, e);
    }
  }

  /**
   */
  private NodeType makeNtFromClass(Class nodeTypeType) throws InstantiationException {

    try {
      Constructor c = nodeTypeType.getConstructor(new Class[] { NodeTypeManager.class });
      return (NodeType) c.newInstance(new Object[] { this });
    } catch (Exception e1) {
      e1.printStackTrace();
      throw new InstantiationException(
          "Error in making istance of "
              + nodeTypeType.getName()
              + ". Class should have been NodeType subclass and have constructor with one argument NodeTypeManager type. Reason: "
              + e1.getCause());
    }
  }

//  public synchronized void initStorage(NodeData sysNode, boolean forcePersistence)
//      throws RepositoryException {
//
//    // 
//    //if (Environment.getInstance().getPlatform() == Environment.STAND_ALONE && !forcePersistence) {
//    //  return;
//    //}
//
//    long start = System.currentTimeMillis();
//    
//    try {
//      persister.initNodetypesRoot(sysNode, forcePersistence);
//      persister.saveChanges();
//      persister.initStorage(nodeTypes);
//      persister.saveChanges();
//      
//    } catch (PathNotFoundException e) {
//      log.error("Error nodetypes storage initialization. " + e.getMessage(), e);
//    }
//    
//    log.info("Nodetypes manager is initialized, initialization time: "
//        + (System.currentTimeMillis() - start) + " ms");
//  }
//
//  public void initStorage(NodeData sysNode) throws RepositoryException {
//    initStorage(sysNode, false);
//  }

  private void initDefault() throws RepositoryException {
    long start = System.currentTimeMillis();
    try {
      InputStream xml = NodeTypeManagerImpl.class.getResourceAsStream(NODETYPES_FILE);
      if (xml != null) {
        registerNodeTypes(xml, ExtendedNodeTypeManager.IGNORE_IF_EXISTS);
      } else {
        String msg = "Resource file '" + NODETYPES_FILE
            + "' with NodeTypes configuration does not found. Can not create node type manager";
        log.error(msg);
        throw new RepositoryException(msg);
      }
    } catch (Exception e) {
      String msg = "Error of initialization default types. Resource file with NodeTypes configuration '" 
        + NODETYPES_FILE + "'. " + e;
      log.error(msg);
      throw new RepositoryException(msg, e);
    } finally {
      log.info("Initialization of default nodetypes done. " + (System.currentTimeMillis() - start) + " ms.");
    }
  }

  
  public void loadFromStorage() throws RepositoryException {
    
    long start = System.currentTimeMillis();
    
    try {
      List<NodeType> loadedNt = persister.loadNodetypes(nodeTypes, this);
      nodeTypes.addAll(loadedNt);
      if (loadedNt.size()>0)
        log.info("NodeTypes (count: " + loadedNt.size() + ") loaded. " + (System.currentTimeMillis() - start) + " ms");
    } catch (PathNotFoundException e) {
      log.warn("NodeTypes storage (/jcr:system/jcr:nodetypes) is not initialized. Only default nodetypes are accessible");
      return;
    }
  }
  

  public boolean isNodeType(InternalQName superTypeName, InternalQName[] typeNames) {

    for (int i = 0; i < typeNames.length; i++) {
      if (superTypeName.equals(typeNames[i])) {
        return true;
      }
      
      NodeType subType;
      NodeType superType;
      try {
        subType = getNodeType(locationFactory.createJCRName(typeNames[i]).getAsString());
        superType = getNodeType(locationFactory.createJCRName(superTypeName).getAsString());
      } catch (RepositoryException e) {
        log.error("Error obtaining node type " + e);
        continue;
      }

      NodeType[] superTypes = subType.getSupertypes();
      for (int j = 0; j < superTypes.length; j++) {
        NodeType testSuperType = superTypes[j];
        if (testSuperType.getName().equals(superType.getName()))
          return true;
      }
    }
    return false;
  }

  public LocationFactory getLocationFactory() {
    return locationFactory;
  }

  public ValueFactory getValueFactory() {
    return valueFactory;
  }
  
///** @deprecated */
//private NodeTypeImpl makeNodeType(NodeData ntNodeData, List<NodeType> ntRegisteredList)
//    throws NoSuchNodeTypeException, ConstraintViolationException, RepositoryException {
//
//  NodeTypeImpl type = new NodeTypeImpl(ntManager);
//      
//  List<PropertyData> nsProps = dataManager.getChildPropertiesData(ntNodeData);
//  boolean jcrNodeTypeName = false;
//  boolean jcrIsMixin = false;
//  boolean jcrHasOrderableChildNodes = false;
//  boolean jcrPrimaryItemName = false;
//  boolean jcrSupertypes = false;
//
//  for (PropertyData ntp : nsProps) {
//    String param = null;
//    List<ValueData> vals = ntp.getValues();
//    if (vals != null && vals.size() > 0) {
//      // param = new String(vals.get(0).getAsByteArray());
//
//      if (ntp.getQPath().getName().equals(Constants.JCR_NODETYPENAME)) { // jcr:nodeTypeName
//        type.setName(valueFactory.loadValue(vals.get(0), PropertyType.STRING).getString());
//        jcrNodeTypeName = true;
//      } else if (ntp.getQPath().getName().equals(Constants.JCR_ISMIXIN)) { // jcr:isMixin
//        type.setMixin(valueFactory.loadValue(vals.get(0), PropertyType.BOOLEAN).getBoolean());
//        jcrIsMixin = true;
//      } else if (ntp.getQPath().getName().equals(Constants.JCR_HASORDERABLECHILDNODES)) { // jcr:hasOrderableChildNodes
//        type.setOrderableChild(valueFactory.loadValue(vals.get(0), PropertyType.BOOLEAN)
//            .getBoolean());
//        jcrHasOrderableChildNodes = true;
//      } else if (ntp.getQPath().getName().equals(Constants.JCR_PRIMARYITEMNAME)) { // jcr:primaryItemName
//        type.setPrimaryItemName(valueFactory.loadValue(vals.get(0), PropertyType.STRING)
//            .getString());
//        jcrPrimaryItemName = true;
//      } else if (ntp.getQPath().getName().equals(Constants.JCR_SUPERTYPES)) { // jcr:supertypes
//        NodeType[] declaredSupertypes = new NodeType[vals.size()];
//        for (int i = 0; i < vals.size(); i++) {
//          String superTypeName = valueFactory.loadValue(vals.get(i), PropertyType.STRING)
//              .getString();
//          for (NodeType regNt : ntRegisteredList) {
//            if (regNt.getName().equals(superTypeName)) {
//              declaredSupertypes[i] = regNt;
//            }
//          }
//          if (declaredSupertypes[i] == null) {
//            throw new ConstraintViolationException("Supertype " + superTypeName
//                + " is not registered before nodetype " + type.getName()
//                + ". Nodetype resistration aborted.");
//          }
//        }
//        type.setDeclaredSupertypes(declaredSupertypes);
//        jcrSupertypes = true;
//      }
//    }
//  }
//  if (jcrNodeTypeName || jcrIsMixin || jcrHasOrderableChildNodes
//      || (jcrSupertypes && !type.isMixin())) {
//    throw new ConstraintViolationException("Mandatory property does not set: "
//        + (jcrNodeTypeName ? "jcr:nodeTypeName" : (jcrIsMixin ? "jcr:isMixin"
//            : (jcrHasOrderableChildNodes ? "jcr:hasOrderableChildNodes" : "jcr:supertypes")))
//        + ". NodeType: " + type.getName());
//  }
////
////  //type.setDeclaredPropertyDefs(getDeclaredPropertyDefs(ntNodeData))
////  //type.setDeclaredNodeDefs(getDeclaredChildNodeDefs(ntNodeData));
//
//  return type;
//}  


//  /** @deprecated never used */
//  private Node makeTypeAsNode(NodeImpl parent, NodeType nodeType) throws PathNotFoundException,
//      RepositoryException, ValueFormatException {
//
//    NodeImpl node = parent.createChildNode(nodeType.getName(), "nt:nodeType", false, true);
//
//    node.setProperty("jcr:nodeTypeName", valueFactory.createValue(nodeType.getName(),
//        PropertyType.NAME));
//    node.setProperty("jcr:isMixin", nodeType.isMixin());
//    node.setProperty("jcr:hasOrderableChildNodes", nodeType.hasOrderableChildNodes());
//    if (nodeType.getPrimaryItemName() != null)
//      node.setProperty("jcr:primaryItemName", nodeType.getPrimaryItemName());
//    Value[] parents = new Value[nodeType.getDeclaredSupertypes().length];
//    for (int i = 0; i < parents.length; i++)
//      parents[i] = valueFactory.createValue(nodeType.getDeclaredSupertypes()[i].getName(),
//          PropertyType.NAME);
//    if (parents.length != 0)
//      node.setProperty("jcr:supertypes", parents);
//    if (nodeType.getDeclaredPropertyDefinitions().length > 0) {
//      for (int i = 0; i < nodeType.getDeclaredPropertyDefinitions().length; i++) {
//        NodeImpl childProps = node.createChildNode("jcr:propertyDefinition",
//            "nt:propertyDefinition", false, true); // same-name sibs
//        initPropertyDefProps(childProps, nodeType.getDeclaredPropertyDefinitions()[i]);
//      }
//    }
//
//    if (nodeType.getDeclaredChildNodeDefinitions().length > 0) {
//      for (int i = 0; i < nodeType.getDeclaredChildNodeDefinitions().length; i++) {
//        Node childNodes = node.createChildNode("jcr:childNodeDefinition", "nt:childNodeDefinition",
//            false, true); // same-name sibs
//        initNodeDefProps(childNodes, nodeType.getDeclaredChildNodeDefinitions()[i]);
//      }
//    }
//    return node;
//  }

//  /** @deprecated */
//  private void initPropertyDefProps(Node node, PropertyDefinition def) throws ValueFormatException,
//      RepositoryException {
//    node.setProperty("jcr:name", valueFactory.createValue(def.getName(), PropertyType.NAME));
//    node.setProperty("jcr:autoCreated", def.isAutoCreated());
//    node.setProperty("jcr:mandatory", def.isMandatory());
//    node.setProperty("jcr:onParentVersion", OnParentVersionAction.nameFromValue(def
//        .getOnParentVersion()));
//    node.setProperty("jcr:protected", def.isProtected());
//    node.setProperty("jcr:requiredType", ExtendedPropertyType.nameFromValue(def.getRequiredType()));
//    node.setProperty("jcr:multiple", def.isMultiple());
//    node.setProperty("jcr:valueConstraints", def.getValueConstraints());
//    node.setProperty("jcr:defaultValues", def.getDefaultValues());
//  }

//  /** @deprecated */
//  private void initNodeDefProps(Node node, NodeDefinition def) throws ValueFormatException,
//      RepositoryException {
//    node.setProperty("jcr:name", valueFactory.createValue(def.getName(), PropertyType.NAME));
//    node.setProperty("jcr:autoCreated", def.isAutoCreated());
//    node.setProperty("jcr:mandatory", def.isMandatory());
//    node.setProperty("jcr:onParentVersion", OnParentVersionAction.nameFromValue(def
//        .getOnParentVersion()));
//    node.setProperty("jcr:protected", def.isProtected());
//    node.setProperty("jcr:sameNameSiblings", def.allowsSameNameSiblings());
//    if (def.getDefaultPrimaryType() != null)
//      node.setProperty("jcr:defaultPrimaryType", def.getDefaultPrimaryType().getName());
//    Value[] reqTypes = new Value[def.getRequiredPrimaryTypes().length];
//    for (int i = 0; i < reqTypes.length; i++)
//      reqTypes[i] = valueFactory.createValue(def.getRequiredPrimaryTypes()[i].getName(),
//          PropertyType.NAME);
//    node.setProperty("jcr:requiredPrimaryTypes", reqTypes, PropertyType.NAME);
//  }


//  /** @deprecated never used */
//  private PropertyDefinition[] getDeclaredPropertyDefs(Node nodeTypeNode)
//      throws NoSuchNodeTypeException, ConstraintViolationException, RepositoryException {
//
//    NodeIterator propertyDefs = nodeTypeNode.getNodes("jcr:propertyDefinition");
//    PropertyDefinition[] declaredPropertyDefs = new PropertyDefinition[(int) propertyDefs.getSize()];
//    int i = 0;
//    while (propertyDefs.hasNext()) {
//      Node propertyDefNode = propertyDefs.nextNode();
//      PropertyDefinitionImpl pDef = new PropertyDefinitionImpl();
//      // log.info("Prop def node "+propertyDefNode.getPath());
//
//      try {
//        pDef.setName(propertyDefNode.getProperty("jcr:name").getString());
//        pDef.setAutoCreate(propertyDefNode.getProperty("jcr:autoCreated").getBoolean());
//        pDef.setMandatory(propertyDefNode.getProperty("jcr:mandatory").getBoolean());
//        pDef.setReadOnly(propertyDefNode.getProperty("jcr:protected").getBoolean());
//        pDef.setMultiple(propertyDefNode.getProperty("jcr:multiple").getBoolean());
//        pDef.setOnVersion(OnParentVersionAction.valueFromName(propertyDefNode.getProperty(
//            "jcr:onParentVersion").getString()));
//        pDef.setDeclaringNodeType(propertyDefNode.getPrimaryNodeType());
//        pDef.setRequiredType(ExtendedPropertyType.valueFromName(propertyDefNode.getProperty(
//            "jcr:requiredType").getString()));
//      } catch (PathNotFoundException e) {
//        throw new ConstraintViolationException("Mandatory property did not set " + e);
//      }
//
//      try {
//        Value[] valueConstraintValues = propertyDefNode.getProperty("jcr:valueConstraints")
//            .getValues();
//        String[] valueConstraints = new String[valueConstraintValues.length];
//        for (int j = 0; j < valueConstraintValues.length; j++) {
//          if (valueConstraintValues[j] != null)
//            valueConstraints[i] = valueConstraintValues[j].getString();
//          else
//            valueConstraints[j] = null;
//        }
//        pDef.setValueConstraints(valueConstraints);
//      } catch (PathNotFoundException e) {
//      }
//      try {
//        pDef.setDefaultValues(propertyDefNode.getProperty("jcr:defaultValues").getValues());
//      } catch (PathNotFoundException e) {
//      }
//
//      declaredPropertyDefs[i++] = pDef;
//    }
//
//    return declaredPropertyDefs;
//
//  }

//  /** @deprecated never used */
//  private NodeDefinition[] getDeclaredChildNodeDefs(Node nodeTypeNode)
//      throws NoSuchNodeTypeException, ConstraintViolationException, RepositoryException {
//
//    NodeIterator childNodeDefs = nodeTypeNode.getNodes("jcr:childNodeDefinition");
//    NodeDefinition[] declaredNodeDefs = new NodeDefinition[(int) childNodeDefs.getSize()];
//    int i = 0;
//    while (childNodeDefs.hasNext()) {
//
//      Node nodeDefNode = childNodeDefs.nextNode();
//      NodeDefinitionImpl nDef = new NodeDefinitionImpl();
//
//      try {
//        nDef.setName(nodeDefNode.getProperty("jcr:name").getString());
//        // nDef.setDefaultNodeType(getNodeType(nodeDefNode.
//        // getProperty("jcr:defaultPrimaryType").getString()));
//
//        Value[] requiredNodeTypesValues = nodeDefNode.getProperty("jcr:requiredPrimaryTypes")
//            .getValues();
//        NodeType[] requiredNodeTypes = new NodeType[requiredNodeTypesValues.length];
//        for (int j = 0; j < requiredNodeTypesValues.length; j++) {
//          if (requiredNodeTypesValues[j] != null)
//            requiredNodeTypes[j] = getNodeType(requiredNodeTypesValues[j].getString());
//          else
//            throw new ConstraintViolationException("Required node type is null ");
//        }
//        nDef.setRequiredNodeTypes(requiredNodeTypes);
//        nDef.setDeclaringNodeType(nodeDefNode.getPrimaryNodeType());
//
//        nDef.setAutoCreate(nodeDefNode.getProperty("jcr:autoCreated").getBoolean());
//        nDef.setMandatory(nodeDefNode.getProperty("jcr:mandatory").getBoolean());
//        nDef.setReadOnly(nodeDefNode.getProperty("jcr:protected").getBoolean());
//        nDef.setMultiple(nodeDefNode.getProperty("jcr:sameNameSiblings").getBoolean());
//        nDef.setOnVersion(OnParentVersionAction.valueFromName(nodeDefNode.getProperty(
//            "jcr:onParentVersion").getString()));
//      } catch (PathNotFoundException e) {
//        throw new ConstraintViolationException("Mandatory property did not set " + e);
//      }
//
//      try {
//        nDef.setDefaultNodeType(getNodeType(nodeDefNode.getProperty("jcr:defaultPrimaryType")
//            .getString()));
//      } catch (PathNotFoundException e) {
//      }
//
//      declaredNodeDefs[i++] = nDef;
//    }
//
//    return declaredNodeDefs;
//
//  }
  
}