/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.util.Calendar;
import java.util.Date;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;


/**
 * A <code>DoubleValue</code> provides an implementation
 * of the <code>Value</code> interface representing a double value.
 *
 * @author Stefan Guggisberg
 */
public class DoubleValue extends BaseValue {

    public static final int TYPE = PropertyType.DOUBLE;

    private final Double dblNumber;


    /**
     * Constructs a <code>DoubleValue</code> object representing a double.
     *
     * @param dbl the double this <code>DoubleValue</code> should represent
     */
    public DoubleValue(double dbl) {
		super(TYPE, new TransientValueData(dbl));
		this.dblNumber = new Double(dbl);
	}

    DoubleValue(ValueData data) {
		super(TYPE, data);
		this.dblNumber = new Double(new String(data.getAsByteArray()));
	}

    /**
     * Indicates whether some other object is "equal to" this one.
     * <p/>
     * The result is <code>true</code> if and only if the argument is not
     * <code>null</code> and is a <code>DoubleValue</code> object that
     * represents the same value as this object.
     *
     * @param obj the reference object with which to compare.
     * @return <code>true</code> if this object is the same as the obj
     *         argument; <code>false</code> otherwise.
     */
    public boolean equals(Object obj) {
	if (this == obj) {
	    return true;
	}
	if (obj instanceof DoubleValue) {
	    DoubleValue other = (DoubleValue) obj;
	    if (dblNumber == other.dblNumber) {
		return true;
	    } else if (dblNumber != null && other.dblNumber != null) {
		return dblNumber.equals(other.dblNumber);
	    }
	}
	return false;
    }

    /**
     * @see BaseValue#getInternalString()
     */
    protected String getInternalString() throws ValueFormatException {
	if (dblNumber != null) {
	    return dblNumber.toString();
	} else {
	    throw new ValueFormatException("empty value");
	}
    }

    //----------------------------------------------------------------< Value >
    /**
     * @see Value#getDate
     */
    public Calendar getDate() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	if (dblNumber != null) {
	    // loosing timezone information...
	    Calendar cal = Calendar.getInstance();
	    cal.setTime(new Date(dblNumber.longValue()));
	    return cal;
	} else {
	    throw new ValueFormatException("empty value");
	}
    }

    /**
     * @see Value#getLong
     */
    public long getLong() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	if (dblNumber != null) {
	    return dblNumber.longValue();
	} else {
	    throw new ValueFormatException("empty value");
	}
    }

    /**
     * @see Value#getBoolean
     */
    public boolean getBoolean() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	throw new ValueFormatException("conversion to boolean failed: inconvertible types");
    }

    /**
     * @see Value#getDouble
     */
    public double getDouble() throws ValueFormatException, IllegalStateException, RepositoryException {
	setValueConsumed();

	if (dblNumber != null) {
	    return dblNumber.doubleValue();
	} else {
	    throw new ValueFormatException("empty value");
	}
    }
}
