/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core.version;

import java.util.Stack;

import javax.jcr.ItemExistsException;
import javax.jcr.ItemNotFoundException;
import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.Property;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.util.TraversingItemVisitor;
import javax.jcr.version.OnParentVersionAction;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.ItemChangeState;
import org.exoplatform.services.jcr.impl.core.ItemState;
import org.exoplatform.services.jcr.impl.core.NodeImpl;
import org.exoplatform.services.jcr.impl.core.PropertyImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.core.itemvisitors.NodeCopyVisitor;
import org.exoplatform.services.jcr.impl.dataflow.MutableItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.TransientNodeData;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL        .
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: RestoreVisitor.java 6386 2006-06-22 16:11:45Z peterit $
 */

class RestoreVisitor extends TraversingItemVisitor  {
  
  protected boolean removeExisting;
  protected Stack<NodeImpl> parents;
  protected NodeImpl context;
  
  protected static Log log = ExoLogger.getLogger("jcr.RestoreVisitor");
  
  RestoreVisitor(NodeImpl context, boolean removeExisting) throws RepositoryException {
  
    super();
    this.parents = new Stack<NodeImpl>();
    this.context = context;
    parents.push(context);
    this.removeExisting = removeExisting; 
  }
  
  protected void toLog(String message) {
    //log.info(message);
  }
  
  protected void entering(Property property, int level) throws RepositoryException {

    int action = property.getDefinition().getOnParentVersion();
    
    // [PN] 10.04.06 In case of COPY, VERSION - copy property, otherwise we don't
    // 8.2.11.3 INITIALIZE; 8.2.11.4 COMPUTE 
    // On restore of VN, the P stored as its child will be ignored, and the
    // current P in the workspace will be left unchanged.
    // [PN] 17.04.06
    //if (action == OnParentVersionAction.COPY || action == OnParentVersionAction.VERSION) {
    if (curParent() != null) {
      toLog("Visit property " + property.getPath()+" "+curParent().getPath() + " " + OnParentVersionAction.nameFromValue(action));
      PropertyInfo info = new PropertyInfo(property);
      String propertyName = info.getName();
      if (propertyName != null) { // ignore if so
        // [PN] 09.03.06
        Property p = createChildProperty(curParent(), info.property, 
            info.getValues(), propertyName, info.getType());
        if (p != null)
          toLog("Restore property " + p.getPath() + " " + curParent().getPath() + " " + curParent().getPrimaryNodeType().getName());
        else 
          toLog("Property '" + propertyName + "' can't be restored at parent " + curParent().getPath() + " " + OnParentVersionAction.nameFromValue(action));
      }
    } else {
      toLog("Visit property " + property.getPath()+" HAS NULL PARENT " + action);
    }
    //}
  }
  
  protected void entering(Node node, int level) throws RepositoryException {
    
    if(node.isNodeType("nt:frozenNode") || curParent() == null) {
      // can be either root or referenced from versionedChild
      toLog("frozenNode "+node.getPath());
    } else if(node.isNodeType("nt:versionedChild")) {
      // i.e. action == OnParentVersionAction.VERSION
      String ref = node.getProperty("jcr:childVersionHistory").getString();
      //VersionHistoryImpl history = versionStorage.getVersionHistoryByUUID(ref, (SessionImpl) node.getSession());
      VersionHistoryImpl history = (VersionHistoryImpl)((SessionImpl)node.getSession()).getNodeByUUID(ref);
      String versionableUuid = history.getVersionableUUID();
      
      toLog("versionedChild node >> "+node.getPath()+" "+versionableUuid);
      
      try {
        NodeImpl childNode = (NodeImpl)history.getSession().getNodeByUUID(versionableUuid);
        toLog("versionable child found >> "+childNode.getPath());
        if(removeExisting) {
//          String name = childNode.getName();
//          String ptName = childNode.getPrimaryNodeType().getName();
//          String[] mtNames = childNode.getMixinTypeNames();
//          NodeImpl newNode = curParent().createChildNode(name, ptName, mtNames, false);
          
          // [PN] 16.02.05
          NodeImpl newNode = createChildNode(curParent(), childNode, removeExisting);

          toLog("copy node >> parent: "+curParent().getPath());

          NodeCopyVisitor copyVisitor = new NodeCopyVisitor(newNode, false, false);
          childNode.accept(copyVisitor);
          
          parents.push(newNode); // [PN] 24.02.06 -- newNode
          
        } else {
          // 2. exists and not a child and removeEx = false (ItemExistsException)
          throw new ItemExistsException("Node "+versionableUuid+" already exists "+childNode.getPath()+" and removeExisting==false"); 
        }
        
      } catch (ItemNotFoundException e) {
        toLog("versionable UUID not found >> "+versionableUuid+" "+node.getPath());
        // 3. not exists
        // gets last version (by time of creation)
        NodeImpl frozen = (NodeImpl)history.getLastVersion().getNode("jcr:frozenNode");
        
        //String name = node.getName();
        String ptName = frozen.getProperty("jcr:frozenPrimaryType").getString();
        Value[] mtValues = frozen.getProperty("jcr:frozenMixinTypes").getValues();
        String[] mtNames = new String[mtValues.length];
        for(int i=0; i<mtNames.length; i++) {
          mtNames[i] = mtValues[i].getString();
        }
        String uuid = frozen.getProperty("jcr:frozenUuid").getString(); // [PN] 26.02.06 did node
        //NodeImpl restoredNode = curParent().createChildNode(name, ptName, mtNames, uuid, true, false);
        
        // [PN] 16.02.06
        NodeImpl nodeImpl = (NodeImpl) node;
        SessionImpl nodeSession = (SessionImpl) node.getSession();
        TransientNodeData nodeData = (TransientNodeData) nodeImpl.getActualItemData();
        InternalQName nodeQName = nodeData.getQName();
        InternalQName ptQName = nodeSession.getLocationFactory().parseJCRName(ptName).getInternalName();
        InternalQName[] mtQNames = new InternalQName[mtNames.length];        
        for (int i=0; i<mtNames.length; i++) {
          String mtName = mtNames[i];
          mtQNames[i] = nodeSession.getLocationFactory().parseJCRName(mtName).getInternalName();
        }
        NodeImpl restoredNode = curParent().createChildNodeInmemory(nodeQName, ptQName, mtQNames, 
            uuid, nodeImpl.getIndex(), false, false, false);
        
        toLog("frozen child node>> "+node.getPath()+" "+node.getPrimaryNodeType().getName());

//        RestoreVisitor childVisitor = new RestoreVisitor(restoredNode, 
//            removeExisting, versionStorage);
        RestoreVisitor childVisitor = new RestoreVisitor(restoredNode, removeExisting); 
        frozen.accept(childVisitor);
        
        parents.push(restoredNode);
      } 
            
    } else { // ordinary node for copy under nt:frozenNode      
      
      // [PN] 10.04.06 In case of COPY - copy node, otherwise we don't
      // 8.2.11.3 INITIALIZE; 8.2.11.4 COMPUTE 
      // On restore of VN, the C stored as its child will be ignored, and the
      // current C in the workspace will be left unchanged.
      toLog("frozen child node>> "+node.getPath()+" "+node.getPrimaryNodeType().getName());
      NodeImpl newNode = createChildNode(curParent(), (NodeImpl) node, removeExisting);
      if (newNode != null) {
        toLog("frozen child node restored>> "+newNode.getPath()+" "+newNode.getPrimaryNodeType().getName());
        parents.push(newNode);
      }
    }
  }
  
  private NodeImpl createChildNode(NodeImpl nodeParent, NodeImpl node, boolean removeExistingNode) throws RepositoryException {
    
    SessionImpl contextSession = (SessionImpl) context.getSession();
    
    boolean isReferenceable = node.isNodeType("mix:referenceable");
    TransientNodeData nodeData = (TransientNodeData) node.getActualItemData();
    InternalQName nodeName = nodeData.getQName();
    InternalQName ptName = nodeData.getPrimaryTypeName();
    InternalQName[] mtNames = nodeData.getMixinTypeNames();
    
    NodeImpl sameUUIDNode = null;
    NodeImpl nonRefSameUUIDNode = null;
    String uuid = null;
    if (isReferenceable) {
      // referenceable node with UUID == jcr:uuid
      uuid = node.getProperty("jcr:uuid").getString();
      // [PN] 13.04.06 look for the same UUID, except nodes in version storage
      try {
        NodeImpl cn = (NodeImpl) contextSession.getNodeByUUID(uuid);
        if(!cn.getActualItemData().getQPath().isDescendantOf(Constants.JCR_VERSION_STORAGE_PATH, false)) {
          sameUUIDNode = cn;
        }
      } catch (ItemNotFoundException e) {
      }      
    } else {
      // [PN] 18.04.06 non referenceable node with UUID == internalUUID, i.e. located by path.
      // [PN] 18.04.06 ??? I'm not sure
//      try {
//        NodeImpl cn = (NodeImpl) contextSession.getItem(node.getPath());
//        //if(!cn.getActualItemData().getQPath().isDescendantOf(Constants.JCR_VERSION_STORAGE_PATH, false)) {
//        nonRefSameUUIDNode = cn;
//        //}
//      } catch (ItemNotFoundException e) {
//      }
    }
    
    // [PN] 13.04.06, 18.04.06
    if (sameUUIDNode != null) {
      if (removeExistingNode) {
        // we have to remove sameUuid node if one new is referenceable
        // [PN] 09.06.06
        //sameUUIDNode.remove();
        //sameUUIDNode.save();
        sameUUIDNode.removeFromParent();
        MutableItemDataChangesLog changeLog = new MutableItemDataChangesLog(contextSession);
        changeLog.add(ItemState.createDeleted(sameUUIDNode.getActualItemData()));
        contextSession.getTransientNodesManager().getTransactManager().save(changeLog); // actual delete
      } else {
        throw new ItemExistsException("Item with the same UUID as " + nodeParent.getPath() + "/" + nodeName.getName() 
            + " already exists and removeExisting = false " + sameUUIDNode.getPath() + " " + uuid);
      }
    }
        
    NodeImpl newNode = nodeParent.createChildNodeInmemory(nodeName, ptName, mtNames, 
        uuid, node.getIndex(), false, false, true); // [PN] 11.04.06 addAutoCreated=true 
    
//    toLog("====== createChildNode() ws: " + contextSession.getWorkspace().getName() 
//        + " src:" + node.getPath()
//        + " node:" + newNode.getPath() + " " + uuid);
       
    // [PN] 11.04.06 
    int action = newNode.getDefinition().getOnParentVersion();
    if (action != OnParentVersionAction.COPY && action != OnParentVersionAction.VERSION) {
      // remove created node from parent
      toLog("===== node can't be restored " + newNode.getPath() + " " + OnParentVersionAction.nameFromValue(action));
      newNode.removeFromParent(ItemChangeState.UNCHANGED);
      return null; // node can't be restored according the OnParentVersionAction rules 
    }

    // [PN] 18.04.06 ??? I'm not sure
//    if (nonRefSameUUIDNode != null) {
//      // for non referenceable node restore
//      newNode.removeFromParent(ItemChangeState.UNCHANGED);
//      nonRefSameUUIDNode.loadData(newNode.getActualItemData());
//    }
    
    return newNode;
  }
  
  private PropertyImpl createChildProperty(NodeImpl propertyParent,
      Property property, Value[] value, String propertyName, int type) throws RepositoryException {

    PropertyImpl propertyImpl = (PropertyImpl) property;
    //SessionImpl propertySession = (SessionImpl) propertyImpl.getSession();
    //InternalQName qname = propertySession.getLocationFactory().parseJCRName(propertyName).getInternalName();
    int action = 0;
    try {
      PropertyImpl existedProperty = (PropertyImpl) propertyParent.getProperty(propertyName);
      action = existedProperty.getDefinition().getOnParentVersion();
      if (action>0 && action != OnParentVersionAction.COPY && action != OnParentVersionAction.VERSION) {
        toLog("Property will be left unchanged. Existed: " + existedProperty.getPath() + " " + OnParentVersionAction.nameFromValue(action)); 
        return null;
      }
    } catch(PathNotFoundException e) {
    }
    // restoring 
    PropertyImpl restoredProperty = propertyParent.createChildPropertyInmemory(propertyName, 
          type, value, propertyImpl.isMultiValued(), false, false);
    action = restoredProperty.getDefinition().getOnParentVersion();
    if (action == OnParentVersionAction.COPY || action == OnParentVersionAction.VERSION) {
      return restoredProperty;
    }
    restoredProperty.removeFromParent(ItemChangeState.UNCHANGED);
    //propertyParent.loadChildProperties().remove(restoredProperty); // [PN] 17.04.06 TODO magic, must be removed
    return null; // property can't be restored according the OnParentVersionAction rules
  }
  
  protected void leaving(Property property, int level)
      throws RepositoryException {
  }

  protected void leaving(Node node, int level) throws RepositoryException {
    parents.pop();
  }
  
  private NodeImpl curParent() {
    return parents.peek();
  }
  
  
  private class PropertyInfo {
    
    private Property property;
    
    private PropertyInfo(Property property) {
      this.property = property;
    }
    
    private String getName() throws RepositoryException {
      String name = property.getName();
      if (property.getParent().isNodeType("nt:frozenNode")
          && (name.equals("jcr:primaryType") 
              || name.equals("jcr:mixinTypes")
              || name.equals("jcr:uuid") 
              || name.equals("jcr:frozenPrimaryType")
              || name.equals("jcr:frozenMixinTypes") 
              || name.equals("jcr:frozenUuid"))
          || name.equals("jcr:childVersionHistory")) {
        // ignore as it belongs to frozen
        return null;
      }
      return name;
    }
    
    private Value[] getValues() throws RepositoryException {
      return ((PropertyImpl)property).getValueArray();
    }
    
    private int getType() throws RepositoryException {
      return property.getType();
    }

  }
}
