
/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core.version;

import java.util.ArrayList;
import java.util.List;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.NodeIterator;
import javax.jcr.PathNotFoundException;
import javax.jcr.Property;
import javax.jcr.PropertyIterator;
import javax.jcr.PropertyType;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.RepositoryException;
import javax.jcr.UnsupportedRepositoryOperationException;
import javax.jcr.Value;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.version.Version;
import javax.jcr.version.VersionException;
import javax.jcr.version.VersionHistory;
import javax.jcr.version.VersionIterator;

import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.ItemChangeState;
import org.exoplatform.services.jcr.impl.core.NodeImpl;
import org.exoplatform.services.jcr.impl.core.PropertyImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.dataflow.TransientNodeData;
import org.exoplatform.services.jcr.impl.util.EntityCollection;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov
 *         </a>
 * @version $Id: VersionHistoryImpl.java 6386 2006-06-22 16:11:45Z peterit $
 */

public class VersionHistoryImpl extends VersionStorageDescendantNode 
  implements VersionHistory {

  /**
   * New constructor
   * @param id
   * @param session
   */
  public VersionHistoryImpl(String id, SessionImpl session) {
    super(id, session);
  }  

  public VersionHistoryImpl(NodeData data, NodeImpl parent, int state,
      SessionImpl session) throws PathNotFoundException,
      RepositoryException {
    super(data, parent, state, session);
    if (!this.getPrimaryNodeType().isNodeType("nt:versionHistory"))
      throw new RepositoryException("Node " + getLocation().getAsString(true)
          + " is not nt:versionHistory type");
  }

  public String getVersionableUUID() throws RepositoryException {
    return this.getProperty("jcr:versionableUuid").getString();
  }

  public Version getRootVersion() throws RepositoryException {
    return (Version) this.getNode("jcr:rootVersion");
  }

  public VersionIterator getAllVersions() throws RepositoryException {
    
    reloadVersionData();
    
    EntityCollection versions = new EntityCollection();
    // should be first in list
    versions.add(getNode("jcr:rootVersion"));

    NodeIterator nodes = getNodes();
    while (nodes.hasNext()) {
      NodeImpl node = (NodeImpl) nodes.nextNode();
      if (!node.getName().equals("jcr:rootVersion")
          && !node.getName().equals("jcr:versionLabels")) {
        //node = new VersionImpl(node.storedNodeData(), this, this.state,
        //    this.session, this.logService);
        versions.add(node);
        //        versions.add(new VersionImpl(node));
      }
    }
    return versions;
  }

  public Version getVersion(String versionName) throws VersionException, RepositoryException {
    
    NodeImpl node;
    try {
      node = (NodeImpl)getNode(versionName);
    } catch (PathNotFoundException e) {
      throw new VersionException("There are no version with name '" + versionName
          + "' in the version history "+getPath());
    }
    
    //return new VersionImpl(node.storedNodeData(), this, this.state,
    //    this.session, this.logService);
    return (Version) node;
  }

  public Version getVersionByLabel(String label) throws RepositoryException {

    String uuid = getUuidByLabel(label);
    if (uuid == null)
      throw new RepositoryException("No label " + label);

    NodeIterator children = this.getNodes();
    while (children.hasNext()) {
      NodeImpl node = (NodeImpl)children.nextNode();
      // version only
      if (node.isNodeType("nt:version") &&
           node.getInternalUUID().equals(uuid)) 
        //return new VersionImpl(node.storedNodeData(), this, this.state,
        //    this.session, this.logService);
        return (Version)node;
    }

    throw new RepositoryException("No version for label " + label);

  }

  public boolean hasVersionLabel(String label) throws RepositoryException {
    if (this.getUuidByLabel(label) == null)
      return false;
    else
      return true;
  }

  public boolean hasVersionLabel(Version version, String label)
      throws VersionException, RepositoryException {

    if (version.getUUID().equals(getUuidByLabel(label)))
      return true;
    else
      return false;
  }

  public String[] getVersionLabels() throws RepositoryException {
    
    PropertyIterator labels = getNode("jcr:versionLabels").getProperties();
    ArrayList labelStrs = new ArrayList();
    while (labels.hasNext()) {
      Property labelProperty = labels.nextProperty();
      if(labelProperty.getType() == PropertyType.REFERENCE) {
        labelStrs.add(labelProperty.getName());
      }
    }
    String[] res = new String[labelStrs.size()];
    for(int i=0; i<labelStrs.size(); i++)
      res[i] = (String)labelStrs.get(i);
    return res;
  }

  public String[] getVersionLabels(Version version) throws VersionException, RepositoryException {
    
    if(!isVersionBelongToThis(version))  
      throw new VersionException("No version "+version.getPath()+" in history "+getPath());

    PropertyIterator labels = getNode("jcr:versionLabels").getProperties();
    ArrayList labelStrs = new ArrayList();
    while (labels.hasNext()) {
      Property labelProperty = labels.nextProperty();
      if (labelProperty.getType() == PropertyType.REFERENCE && 
          labelProperty.getString().equals(version.getUUID()))
        labelStrs.add(labelProperty.getName());
    }
    String[] res = new String[labelStrs.size()]; 
    for(int i=0; i<labelStrs.size(); i++)
      res[i] = (String)labelStrs.get(i);
    return res;
  }

  public void removeVersion(String versionName)
      throws ReferentialIntegrityException, AccessDeniedException,
      UnsupportedRepositoryOperationException, VersionException,
      RepositoryException {
    VersionImpl ver = (VersionImpl)getVersion(versionName);

    // check references. 
    // Note: References from /jcr:system/jcr:versionStorage never included to getReferences!
    List refs = workspaceDataManager().getReferencesData(ver.getUUID());
    if(refs.size() > 0)
      throw new ReferentialIntegrityException(
          "There are Reference property pointed to this Version "+((PropertyData)refs.get(0)).getQPath().getAsString());

    // [PN] 09.04.06
    //----------- BEGIN SESSION BATCH ----------      
    workspaceDataManager().beginSessionBatch(session);
    
    // remove labels first
    PropertyIterator labels = getNode("jcr:versionLabels").getProperties();
    while (labels.hasNext()) {
      PropertyImpl labelProperty = (PropertyImpl) labels.nextProperty();
      if (labelProperty.getString().equals(ver.getUUID())) {
        labelProperty.removeFromParent();
        session.getTransientNodesManager().getTransactManager().saveItem(labelProperty);
      }
    }
    
    // remove this version from successor anf predecessor list
    // and point successor to predecessor directly
    Version[] predess = ver.getPredecessors();
    Version[] success = ver.getSuccessors();
    for(int i=0; i<predess.length; i++) {
      //log.info("predess >> "+predess[i].getPath()+" "+predess[i].getUUID());
      ((VersionImpl)predess[i]).removeRefValue(true, ver.getUUID());
      for(int j=0; j<success.length; j++) { 
        //log.info("addSuccessor >> "+success[j].getPath()+" "+success[j].getUUID());
        ((VersionImpl)predess[i]).addSuccessor(success[j]);
      }
      //log.info("predess << ");
    }
    
    for(int i=0; i<success.length; i++) {
      //log.info("success >> "+success[i].getPath()+" "+success[i].getUUID());
      ((VersionImpl)success[i]).removeRefValue(false, ver.getUUID());
      for(int j=0; j<predess.length; j++) {
        //log.info("addPredecessor >> "+predess[j].getPath()+" "+predess[j].getUUID());
        ((VersionImpl)success[i]).addPredecessor(predess[j]);
      }
      //log.info("success << ");
    }
    
    ver.removeFromParent();
    session.getTransientNodesManager().getTransactManager().saveItem(ver);
    workspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------      
    session.getTransientNodesManager().resetItemState(ver);
    session.getTransientNodesManager().removeFromCache(ver);
  }

  public void addVersionLabel(String versionName, String label, boolean moveLabel) throws VersionException, RepositoryException {
        Version ver = getVersion(versionName);

    if (hasVersionLabel(label)) {
      if (moveLabel) {
        removeVersionLabel(label);
      } else {
        throw new VersionException("Label " + label + " already exists and moveLabel == false");
      }
    }
    
    NodeImpl versionLabels = (NodeImpl) getNode("jcr:versionLabels");
    Value[] values = new Value[] { valueFactory.createValue(ver) };
    PropertyImpl prop = versionLabels
        .createChildProperty(label, values, PropertyType.REFERENCE);    
    
    // [PN] 17.04.06
    //----------- BEGIN SESSION BATCH ----------
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);    
    dataManager.getTransactManager().saveItem(prop);
    dataManager.getWorkspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------
    dataManager.resetItemState(prop);
    
    //----------- SESSION CACHING ------------
    //dataManager.cacheItem(prop, true);
  }

  public void removeVersionLabel(String label) throws VersionException,
      RepositoryException {

    PropertyImpl prop;
    try {
      prop = (PropertyImpl) getNode("jcr:versionLabels").getProperty(
          label);
    } catch (PathNotFoundException e) {
      throw new VersionException("Label not found " + label);
    } 
    //prop.doRemove(false);
    //prop.save();
    // [PN] 17.04.06
    //----------- BEGIN SESSION BATCH ----------
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);
    prop.removeFromParent();
    dataManager.getTransactManager().saveItem(prop);
    dataManager.getWorkspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------
    dataManager.resetItemState(prop);
  }

  ////////////////// impl

  public VersionImpl addVersion(NodeImpl versionableNode) throws RepositoryException {

    String uuid = uuidGenerator.generateUUID(versionableNode.getPath()).getString();
    VersionImpl version = createVersion(nextVersionName(), uuid);
    
    version.createChildProperty("jcr:primaryType", new Value[] { valueFactory
        .createValue("nt:version", PropertyType.NAME) }, PropertyType.NAME);
    version.createChildProperty("jcr:mixinTypes", new Value[] { valueFactory
        .createValue("mix:referenceable", PropertyType.NAME) }, PropertyType.NAME);
    version.createChildProperty("jcr:uuid", new Value[] { valueFactory
        .createValue(version.storedNodeData().getUUID()) },
        PropertyType.STRING);
    PropertyImpl pc = version.createChildProperty("jcr:created", new Value[] { valueFactory
        .createValue(workspaceDataManager().getCurrentTime()) }, PropertyType.DATE);
//    log.info("addVersion(): " + pc.getDefinition().getName()
//      + " " + OnParentVersionAction.nameFromValue(pc.getDefinition().getOnParentVersion()));

    Value[] predecessors = versionableNode.getProperty("jcr:predecessors").getValues();
    
    //N�s current jcr:predecessors property is copied to V
    version.createChildProperty("jcr:predecessors", predecessors, PropertyType.REFERENCE);

//    RepositoryImpl repo = (RepositoryImpl) session.getRepository();
//    String[] wsNames = repo.getWorkspaceNames();    
//    log.info("===== add version for " + session.getWorkspace().getName() + "  "
//        + versionableNode.getSession().getWorkspace().getName() + ":"
//        + versionableNode.getPath() + "  " + versionableNode.getUUID());
    
    for (int i = 0; i < predecessors.length; i++) {      
      // TODO it is not correct as there can be version added to corresp node in other WS
      String predecessorUuid = predecessors[i].getString();
            
      VersionImpl predec = (VersionImpl) session.getNodeByUUID(predecessorUuid);
      ItemData predecData = session.getTransientNodesManager().getWorkspaceDataManager()
        .getItemData(predecessorUuid);
      predec.loadData(predecData);
      predec.addSuccessor(version);
      // [PN] 12.04.06 Reset change state here, as we can't to know about this change below 
      session.getTransientNodesManager().resetItemState(predec);
    }
    
    // Frozen
    NodeImpl frozen = version.createChildNode("jcr:frozenNode", "nt:frozenNode", true, true);
    versionableNode.accept(new FrozenNodeInitializer(frozen));
    
    return version;
  }
  
  public VersionImpl addRootVersion(NodeImpl versionableNode) throws RepositoryException {

    // Root version
    VersionImpl rootVersion = createVersion("jcr:rootVersion", 
        versionableNode.getProperty("jcr:baseVersion").getString());

    rootVersion.createChildProperty("jcr:primaryType", new Value[] { valueFactory
            .createValue("nt:version", PropertyType.NAME) }, PropertyType.NAME);
    rootVersion.createChildProperty("jcr:mixinTypes", new Value[] { valueFactory
        .createValue("mix:referenceable", PropertyType.NAME) }, PropertyType.NAME);

    rootVersion.createChildProperty("jcr:uuid", new Value[] { valueFactory
        .createValue(versionableNode.getProperty("jcr:baseVersion").getString()) },
        PropertyType.STRING);

    rootVersion.createChildProperty("jcr:created", new Value[] { valueFactory
        .createValue(workspaceDataManager().getCurrentTime()) }, PropertyType.DATE);
    
    return rootVersion;

    // Frozen
//    NodeImpl frozenNode = rootVersion.createChildNode("jcr:frozenNode",
//        "nt:frozenNode", false);
//    frozenNode.createChildProperty("jcr:primaryType",
//        new Value[] { valueFactory.createValue("nt:frozenNode",
//            PropertyType.NAME) }, PropertyType.NAME);
//
//    frozenNode.createChildProperty("jcr:frozenPrimaryType",
//        new Value[] { valueFactory.createValue(versionableNode.getPrimaryNodeType()
//            .getName(), PropertyType.NAME) }, PropertyType.NAME);
//
//    Value[] mixinNodeTypes = new Value[versionableNode.getMixinNodeTypes().length];
//    for (int i = 0; i < mixinNodeTypes.length; i++) {
//      mixinNodeTypes[i] = valueFactory.createValue(versionableNode.getMixinNodeTypes()[i]
//          .getName(), PropertyType.NAME);
//    }
//    frozenNode.createChildProperty("jcr:frozenMixinTypes", mixinNodeTypes,
//        PropertyType.NAME);
//    frozenNode.createChildProperty("jcr:frozenUuid", new Value[] { valueFactory
//        .createValue(versionableNode.getUUID()) }, PropertyType.STRING);
  }  
  
  public boolean isVersionBelongToThis(Version version) {
    return ((VersionImpl)version).getLocation().isDescendantOf(getLocation(), false);
  }
  
  public Version getLastVersion() throws RepositoryException {
    VersionIterator versions = getAllVersions();
    Version last = null;
    while(versions.hasNext()) {
      Version ver = versions.nextVersion();
      if(last == null || ver.getCreated().after(last.getCreated()));
        last = ver;
    }
    return last;
  }
  

  private String getUuidByLabel(String label) throws RepositoryException {
    PropertyIterator labels = getNode("jcr:versionLabels").getProperties();
    while (labels.hasNext()) {
      Property labelProperty = labels.nextProperty();
      if (labelProperty.getName().equals(label))
        return labelProperty.getValue().getString();
    }

    return null;

  }
  
  private VersionImpl createVersion(String versionName, String uuid) throws RepositoryException {
  
    // [PN] 07.06.06
    //JCRPath path = locationFactory.createJCRPath(this.getLocation(), versionName);
    
    NodeData nodeData = TransientNodeData.createNodeData((NodeData) getActualItemData(),
        new InternalQName(null, versionName), Constants.NT_VERSION, uuid);
    
//    InternalQPath path = InternalQPath.makeChildPath(getInternalPath(), "[]" + versionName);
//    int orderNum = 0;
//    NodeData nodeData = new TransientNodeData(path, uuid, -1,
//        Constants.NT_VERSION, new InternalQName[0], orderNum, 
//        storedNodeData().getUUID()); 
    
    VersionImpl version = new VersionImpl(nodeData, this, ItemChangeState.ADDED, session);
    addChildNode(version);
    return version;
  }
  
  private void reloadVersionData() throws InvalidItemStateException, ConstraintViolationException, RepositoryException  {
    //ItemData newData = dataManager.getWorkspaceDataManager().getItemData(getLocation());
    ItemData newData = dataManager.getTransactManager().getItemData(getLocation().getInternalPath());
    loadData(newData);
  }
  
  private String nextVersionName() throws RepositoryException {
    // make version name - 1-based (because of rootVersion) integer    
    int vn = 0;
    for (VersionIterator allVersions = getAllVersions(); allVersions.hasNext();) {
      Version v = allVersions.nextVersion();
      try {
        int vi = Integer.parseInt(v.getName());
        if (vi > vn) {
          vn = vi;
        }
      } catch(NumberFormatException e) {
      }      
    }
    //return new Integer((int) getAllVersions().getSize()).toString();
    return vn > 0 ? String.valueOf(vn + 1) : "1";
  }
}