/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.core.version;

import java.util.ArrayList;
import java.util.List;

import javax.jcr.ItemNotFoundException;
import javax.jcr.PathNotFoundException;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.ItemChangeState;
import org.exoplatform.services.jcr.impl.core.ItemState;
import org.exoplatform.services.jcr.impl.core.NodeImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.dataflow.ItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.TransientNodeData;
import org.exoplatform.services.jcr.impl.dataflow.TransientPropertyData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.jcr.impl.dataflow.persistent.WorkspaceDataManager;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL
 * Author : Peter Nedonosko
 *          peter.nedonosko@exoplatform.com.ua
 * 07.06.2006
 * @version $Id: $
 */
public class VersionHistoryInitializer {
  
  protected static Log log = ExoLogger.getLogger("jcr.VersionHistoryInitializer");
  
  private WorkspaceDataManager dataManager = null;
  private SessionImpl session = null;
  
  public VersionHistoryInitializer(WorkspaceDataManager dataManager, SessionImpl session) {
    this.dataManager = dataManager;
    this.session = session;
  }

  private boolean isVSDescendant(InternalQPath path) {
    return path.equals(Constants.JCR_VERSION_STORAGE_PATH) || path.isDescendantOf(Constants.JCR_VERSION_STORAGE_PATH, false);
  }
  
  private void findVersionable(ItemData item, List<ItemData> versionableChanges, ItemDataChangesLog changesLog) throws RepositoryException {
    
    if (isVSDescendant(item.getQPath())) 
      return; // not applicable for version storage nodes 
    
    if (!item.isNode() && item.getQPath().getName().equals(Constants.JCR_VERSIONHISTORY)) {
      // mix:versionable seems to be assigned
      PropertyData persistedItem = (PropertyData) dataManager.getItemData(item.getQPath());
      if (persistedItem == null) {
        //ItemData itemParent = dataManager.getItemData(item.getParentUUID());
        NodeData parentNode = null;
        try {
          // GA Do not use API's getter here!!!
          // TODO try to find parent in changesLog using stuff like changesLog.getState(String id) 
          parentNode = (NodeData) ((NodeImpl) changesLog.getSession().getNodeByUUID(item.getParentUUID())).getActualItemData();
        } catch(ItemNotFoundException e) {
          // search parent node in changes log if in session we haven't it
          parentNode = (NodeData) changesLog.findItemState(item.getParentUUID()).getData();
          if (parentNode == null) 
            throw e;
        }
        versionableChanges.add(parentNode); // add mix:versionable node in list of new versionable nodes
      }
    }
    
    if(item.isNode()) {
      List<PropertyData> itemChildProperties = dataManager.getChildPropertiesData((NodeData) item);
      for (PropertyData pd: itemChildProperties) {
        findVersionable(pd, versionableChanges, changesLog);
      }
      List<NodeData> itemChildNodes = dataManager.getChildNodesData((NodeData) item);
      for (NodeData nd: itemChildNodes) {
        findVersionable(nd, versionableChanges, changesLog);
      }
    }
  }
  
  public void saveVersionable(ItemDataChangesLog changesLog) throws RepositoryException {
    
    List<ItemState> changes = changesLog.getChanges();
    
    if (changesLog.getSession() != null) {
      List<ItemData> versionableNew = new ArrayList<ItemData>();
      for (int i = 0; i < changes.size(); i++) {
        if (changes.get(i).getState() == ItemChangeState.ADDED 
            || changes.get(i).getState() == ItemChangeState.AUTO_ADDED) {
          findVersionable(changes.get(i).getData(), versionableNew, changesLog);
        }
      }
      
      for (ItemData versionableNode: versionableNew) {
        initVersionHistory((NodeData) versionableNode, changes);
      }
    }
  }
  
  private void initVersionHistory(NodeData versionableNode, List<ItemState> changes) throws RepositoryException {
    
    //log.info("ItemData: " + versionableNode.getQPath().getAsString());
    
    String versionHistoryUuid = null;
    String baseVersionUuid = null;

    // looking for required uuids (jcr:versionHistory, jcr:baseVersion) in changes list first.
    // In case of addMixin("mix:versionable") on a new node.
    for (ItemState itemState : changes) {

      if (itemState.getData().isNode())
        continue; // need properties only

      PropertyData prop = (PropertyData) itemState.getData();
      //log.info("  PropertyData: " + prop.getQPath().getAsString());
      if (prop.getQPath().isDescendantOf(versionableNode.getQPath(), true)) {
        if (prop.getQPath().getName().equals(Constants.JCR_VERSIONHISTORY)) {
          List<ValueData> pVals = prop.getValues();
          if (pVals != null && pVals.size() > 0) {
            versionHistoryUuid = ((ValueFactoryImpl) session.getValueFactory()).loadValue(
                pVals.get(0), PropertyType.REFERENCE).getString();
          } else {
            throw new RepositoryException("Property jcr:versionHistory is empty: ["
                + (pVals != null ? "size 0" : "null") + "]. Parent: "
                + versionableNode.getQPath().getAsString());
          }
        } else if (prop.getQPath().getName().equals(Constants.JCR_BASEVERSION)) {
          List<ValueData> pVals = prop.getValues();
          if (pVals != null && pVals.size() > 0) {
            baseVersionUuid = ((ValueFactoryImpl) session.getValueFactory()).loadValue(
                pVals.get(0), PropertyType.REFERENCE).getString();
          } else {
            throw new RepositoryException("Property jcr:baseVersion is empty: ["
                + (pVals != null ? "size 0" : "null") + "]. Parent: "
                + versionableNode.getQPath().getAsString());
          }
        }
      }
    }
    
    if (versionHistoryUuid == null || baseVersionUuid == null) {
      // if some of required uuids not found in changes list try to search direct in versionable node properties.
      // In case of addMixin("mix:versionable") on already existing node.
      for (Object propObj: ((TransientNodeData) versionableNode).getChildProperties()) {
        PropertyData prop = (PropertyData) propObj;
        //log.info("  PropertyData: " + prop.getQPath().getAsString());
        if (versionHistoryUuid == null && prop.getQPath().getName().equals(Constants.JCR_VERSIONHISTORY)) {
          List<ValueData> pVals = prop.getValues();
          if (pVals != null && pVals.size()>0) {
            versionHistoryUuid = ((ValueFactoryImpl) session.getValueFactory()).loadValue(pVals.get(0), 
                PropertyType.REFERENCE).getString();
          } else {
            throw new RepositoryException("Child property jcr:versionHistory is empty: [" +
                (pVals != null ? "size 0" : "null" )+ "]. Parent: " + versionableNode.getQPath().getAsString());
          }
        } else if (baseVersionUuid == null && prop.getQPath().getName().equals(Constants.JCR_BASEVERSION)) {
          List<ValueData> pVals = prop.getValues();
          if (pVals != null && pVals.size()>0) {
            baseVersionUuid = ((ValueFactoryImpl) session.getValueFactory()).loadValue(pVals.get(0), 
                PropertyType.REFERENCE).getString();
          } else {
            throw new RepositoryException("Child property jcr:baseVersion is empty: [" +
                (pVals != null ? "size 0" : "null" )+ "]. Parent: " + versionableNode.getQPath().getAsString());
          }
        }
      }
    }

    if (versionHistoryUuid == null) 
      throw new RepositoryException("Property jcr:versionHistory must exists in node " 
          + versionableNode.getQPath().getAsString());
    if (baseVersionUuid == null)
      throw new RepositoryException("Property jcr:baseVersion must exists in node " 
          + versionableNode.getQPath().getAsString());
    
    // TODO [PN] It's wrong, as we have not enter here if versionHistory already exists
    // Fix it after session refactor according final architecture.
    if (dataManager.getItemData(versionHistoryUuid) != null)
      return; // version history already exists
    
    // ----- version history -----
    NodeData versionStorageData = (NodeData) dataManager.getItemData(Constants.JCR_VERSION_STORAGE_PATH);
    NodeData versionHistory = TransientNodeData.createNodeData(versionStorageData,
        new InternalQName(null, versionableNode.getUUID()), Constants.NT_VERSIONHISTORY, versionHistoryUuid);
    
    //jcr:primaryType
    TransientPropertyData vhPrimaryType = TransientPropertyData.createPropertyData(
        versionHistory, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
    vhPrimaryType.setValue(new TransientValueData(versionHistory.getPrimaryTypeName()));
    
    //jcr:uuid
    TransientPropertyData vhUuid = TransientPropertyData.createPropertyData(
        versionHistory, Constants.JCR_UUID, PropertyType.STRING, false);
    vhUuid.setValue(new TransientValueData(versionHistoryUuid));
    
    //jcr:versionableUuid
    TransientPropertyData vhVersionableUuid = TransientPropertyData.createPropertyData(
        versionHistory, Constants.JCR_VERSIONABLEUUID, PropertyType.REFERENCE, false);
    vhVersionableUuid.setValue(new TransientValueData(versionableNode.getUUID()));
    
    // ------ jcr:versionLabels ------
    NodeData vhVersionLabels = TransientNodeData.createNodeData(versionHistory,
        Constants.JCR_VERSIONLABELS, Constants.NT_VERSIONLABELS);
    
    // jcr:primaryType
    TransientPropertyData vlPrimaryType = TransientPropertyData.createPropertyData(
        vhVersionLabels, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
    vlPrimaryType.setValue(new TransientValueData(vhVersionLabels.getPrimaryTypeName()));
        
    // ------ jcr:rootVersion ------
    NodeData rootVersionData = TransientNodeData.createNodeData(versionHistory, 
        Constants.JCR_ROOTVERSION, Constants.NT_VERSION, baseVersionUuid);
    
    // jcr:primaryType
    TransientPropertyData rvPrimaryType = TransientPropertyData.createPropertyData(
        rootVersionData, Constants.JCR_PRIMARYTYPE, PropertyType.NAME, false);
    rvPrimaryType.setValue(new TransientValueData(rootVersionData.getPrimaryTypeName()));
    
    //jcr:uuid
    TransientPropertyData rvUuid = TransientPropertyData.createPropertyData(
        rootVersionData, Constants.JCR_UUID, PropertyType.STRING, false);
    rvUuid.setValue(new TransientValueData(baseVersionUuid));
    
    // jcr:mixinTypes
    TransientPropertyData rvMixinTypes = TransientPropertyData.createPropertyData(
        rootVersionData, Constants.JCR_MIXINTYPES, PropertyType.NAME, true);
    rvMixinTypes.setValue(new TransientValueData(Constants.MIX_REFERENCEABLE));
    
    //jcr:created
    TransientPropertyData rvCreated = TransientPropertyData.createPropertyData(
        rootVersionData, Constants.JCR_CREATED, PropertyType.DATE, false);
    rvCreated.setValue(new TransientValueData(dataManager.getCurrentTime()));
    
    changes.add(ItemState.createAdded(versionHistory));
    changes.add(ItemState.createAdded(vhPrimaryType));
    changes.add(ItemState.createAdded(vhUuid));
    changes.add(ItemState.createAdded(vhVersionableUuid));
    
    changes.add(ItemState.createAdded(vhVersionLabels));
    changes.add(ItemState.createAdded(vlPrimaryType));
    
    changes.add(ItemState.createAdded(rootVersionData));
    changes.add(ItemState.createAdded(rvPrimaryType));
    changes.add(ItemState.createAdded(rvUuid));
    changes.add(ItemState.createAdded(rvMixinTypes));
    changes.add(ItemState.createAdded(rvCreated));
  }

  
  // -------------- any stuff ---------------
  
  Value readTransientPropertyValue(TransientNodeData transientNode, 
      InternalQName propertyName, int propertyType) throws PathNotFoundException, RepositoryException {
    for (Object propObj: transientNode.getChildProperties()) {
      PropertyData prop = (PropertyData) propObj;
      if (prop.getQPath().getName().equals(propertyName)) {
        List<ValueData> pVals = prop.getValues();
        if (pVals != null && pVals.size()>0) {
          return ((ValueFactoryImpl) session.getValueFactory()).loadValue(pVals.get(0), propertyType);
        }
        throw new RepositoryException("Property '"+propertyName.getAsString()+"' is empty: [" +
            (pVals != null ? "0" : "null" )+ "]. Parent: " + transientNode.getQPath().getAsString());
      }
    }
    throw new PathNotFoundException("Property '" + propertyName.getAsString() 
        + "' not found. Parent: " + transientNode.getQPath().getAsString());
  }
}
